#!/usr/bin/env python3
import contextlib
import errno
import os
import signal

DEFAULT_MESSAGE = os.strerror(errno.ETIME)

class QETimeoutError(Exception):
    def __init__(self, message):
        Exception.__init__(self, message)


class Timeout(contextlib.ContextDecorator):
    def __init__(self, seconds, timeout_message=DEFAULT_MESSAGE, suppress_timeout_errors=False):
        self.seconds = int(seconds)
        self.timeout_message = timeout_message
        self.suppress = bool(suppress_timeout_errors)

    def _timeout_handler(self, signum, frame):
        raise QETimeoutError(self.timeout_message)

    def __enter__(self):
        signal.signal(signal.SIGALRM, self._timeout_handler)
        signal.alarm(self.seconds)

    def __exit__(self, exception_type, exception_valeu, exception_traceback):
        signal.alarm(0)
        if self.suppress and exception_type is QETimeoutError:
            return



#!/usr/bin/env python3
import os
import re
import base64
from time import strftime, sleep, localtime
from subprocess import Popen
from dogtail.tree import root
from dogtail.utils import config
from dogtail.rawinput import keyCombo
from qecore import run, QE_LOGGER
from qecore.application import Application
from qecore.time_limit import Timeout, QETimeoutError


class TestSandbox:
    def __init__(self, component, logging=False):
        """
        Initiate TestSandbox instance.

        @param component : str
            Testing component name.
        @param logging : bool
            Turn on or off logging of this submodule.
            Default option is False.
            Mostly used if submodule encounters a corner case that was not caught in implementation.
        """

        self.logging = logging

        if self.logging:
            QE_LOGGER.info(f"__init__(self, component={component}, logging={str(self.logging)})")

        self.component = component
        self.current_scenario = None
        self.background_color = None

        self.logging_start = None
        self.screenshot_run_result = None

        self.record_video = True
        self.record_video_pid = None

        self.attach_video = True
        self.attach_video_on_pass = False

        self.attach_journal = True
        self.attach_journal_on_pass = False

        self.attach_screenshot = True
        self.failed_test = False

        self.attach_faf = True
        self.attach_faf_on_pass = False

        self.workspace_return = False

        self.set_keyring = True

        self.wait_for_stable_video = True

        self.production = True

        self.applications = []
        self.default_application = None
        self.retrieve_session_data()
        self.wait_until_shell_becomes_responsive()

        self.shell = root.application("gnome-shell")


    def before_scenario(self, context, scenario):
        """
        Actions that are to be executed before every scenario.

        @param context : <context_object>
            Pass this object from environment file.
        @param scenario : <behave_object>
            Pass this object from environment file.
        """

        if self.logging:
            QE_LOGGER.info(f"before_scenario(self, context, scenario) test: {scenario.tags[-1]}")

        self.failed_test = False

        self.current_scenario = scenario.tags[0]
        self.set_log_start_time()
        self.overview_action("hide")
        self.set_typing_delay(0.2)
        self.set_debug_to_stdout_as(False)
        self.close_yelp()
        self.close_initial_setup()
        self.copy_data_folder()
        self.set_blank_screen_to_never()

        self.set_up_embedding(context)

        if self.record_video and self.production:
            self.start_recording()

        self.detect_keyring()
        self.return_to_home_workspace()


    def after_scenario(self, context, scenario):
        """
        Actions that are to be executed after every scenario.

        @param context : <context_object>
            Pass this object from environment file.
        @param scenario : <behave_object>
            Pass this object from environment file.
        """

        if self.logging:
            QE_LOGGER.info(f"after_scenario(self, context, scenario) test: {scenario.tags[0]}")

        if scenario.status == "failed":
            self.failed_test = True

        if self.attach_screenshot and self.failed_test:
            self.capture_image()

        if self.record_video:
            self.stop_recording()

        self.overview_action("hide")

        for application in self.applications:
            application.kill_application()

        if self.attach_screenshot and self.failed_test:
            self.attach_screenshot_to_report(context)

        if self.attach_journal and (self.failed_test or self.attach_journal_on_pass):
            self.attach_journal_to_report(context)

        if self.attach_video and (self.failed_test or self.attach_video_on_pass):
            self.attach_video_to_report(context)

        if self.attach_faf and (self.failed_test or self.attach_faf_on_pass):
            self.attach_abrt_link_to_report(context)


    def start_recording(self):
        """
        Start recording video.
        """

        if self.logging:
            QE_LOGGER.info("start_recording(self)")

        self.display_clock_seconds()
        self.set_max_video_length_to(600)

        record_video_process = Popen("qecore_start_recording", shell=True)
        self.record_video_pid = record_video_process.pid


    def stop_recording(self):
        """
        Stop recording video.
        """

        if self.logging:
            QE_LOGGER.info("stop_recording(self)")

        if self.record_video_pid is not None:
            run(f"sudo kill -9 {self.record_video_pid} > /dev/null")

        self.record_video_pid = None


    def get_app(self, name, a11yAppName=None, desktopFileExists=True,
                desktopFileName=None, appProcessName=None):
        """
        Return application to be used in test.

        @param name : str
            Name of application.
        @param a11yAppName : str
            Name of application in a11y tree.
        @param desktopFileExists : bool
            Does desktop file of application exist.
        @param desktopFileName : str
            Name of desktop file of application.
        @param appProcessName : str
            Name of process after application starts.
        """

        if self.logging:
            QE_LOGGER.info("".join((
                f"get_app(self, name={name}, a11yAppName={a11yAppName}, ",
                f"desktopFileExists={desktopFileExists}, desktopFileName={desktopFileName})"
            )))

        new_application = Application(name, a11yAppName, desktopFileExists,
                                      desktopFileName, appProcessName,
                                      self.shell, self.session_type,
                                      self.session_desktop, self.logging)

        self.applications.append(new_application)
        self.default_application = new_application \
            if self.default_application is None else self.default_application

        return new_application


    def wait_until_shell_becomes_responsive(self):
        """
        Give some time if shell is not yet loaded fully.
        """

        if self.logging:
            QE_LOGGER.info("wait_until_shell_becomes_responsive(self)")

        if self.session_type == "x11":
            with Timeout(30):
                while not "gnome-shell" in [x.name for x in root.applications()]:
                    sleep(0.5)
        else:
            sleep(1) # in wayland there is no way to make sure the session is loaded, so just sleep


    def retrieve_session_data(self):
        """
        Get session type/desktop data.
        """

        if self.logging:
            QE_LOGGER.info("retrieve_session_data(self)")

        self.architecture = run("uname -m").strip("\n")

        # Distributions expected for now: self.distribution = ["Red Hat Enterprise Linux", "Fedora"]
        self.distribution = run("cat /etc/os-release | grep ^NAME=")
        self.distribution = self.distribution.split("=")[-1].strip("\n").strip("\"")

        try:
            self.resolution = [int(x) for x in \
                re.findall(r"\d+x\d+", run("xrandr | grep '*'"))[0].split("x")]
        except Exception as error:
            self.resolution = f"The resolution retrieval failed for: {error}"

        self.session_desktop = run("echo $XDG_SESSION_DESKTOP").strip('\n')
        self.session_type = "x11"
        if "XDG_SESSION_TYPE" in os.environ and "wayland" in os.environ["XDG_SESSION_TYPE"]:
            self.session_type = "wayland"


    def set_up_embedding(self, context):
        def embed_data(mime_type, data, caption):
            for formatter in context._runner.formatters:
                if "html" in formatter.name:
                    self._embed_to = formatter

            if hasattr(self, "_embed_to"):
                self._embed_to.embedding(mime_type=mime_type, data=data, caption=caption)
            else:
                return None

        context.embed = embed_data


    def set_log_start_time(self):
        """
        Save time. Will be used to retrieve logs from journal.
        """

        if self.logging:
            QE_LOGGER.info("set_log_start_time(self)")

        self.logging_start = strftime("%Y-%m-%d %H:%M:%S", localtime())


    def close_yelp(self):
        """
        Close yelp application that is opened after fresh installation.
        """

        if self.logging:
            QE_LOGGER.info("close_yelp(self)")

        help_process_id = run("pgrep yelp").strip("\n")
        if help_process_id.isdigit():
            run(f"kill -9 {help_process_id}")


    def close_initial_setup(self):
        """
        Close initial setup window that is opened after first login.
        """

        if self.logging:
            QE_LOGGER.info("close_initial_setup(self)")

        run("echo yes > ~/.config/gnome-initial-setup-done")


    def set_blank_screen_to_never(self):
        """
        Set blank screen to never. For longer tests it is undesirable for screen to lock.
        """

        if self.logging:
            QE_LOGGER.info("set_blank_screen_to_never(self)")

        run("gsettings set org.gnome.desktop.session idle-delay 0")


    def set_max_video_length_to(self, number=600):
        """
        Set maximum allowed video lenght.

        @param number : int
            Maximum video lenght.
        """

        if self.logging:
            QE_LOGGER.info(f"set_max_video_length_to(self, number={number})")

        run(" ".join((
            f"gsettings set",
            f"org.gnome.settings-daemon.plugins.media-keys",
            f"max-screencast-length {number}"
        )))


    def display_clock_seconds(self):
        """
        Display clock seconds for better tracking test in video.
        """

        if self.logging:
            QE_LOGGER.info("display_clock_seconds(self)")

        run("gsettings set org.gnome.desktop.interface clock-show-seconds true")


    def return_to_home_workspace(self):
        """
        Return to home workspace.
        """

        if self.logging:
            QE_LOGGER.info("return_to_home_workspace(self)")

        if not self.workspace_return:
            return

        keyCombo("<Super><Home>")


    def set_typing_delay(self, number):
        """
        Set typing delay so slower machines will not lose characters on type.

        @param number : int
            Time in between accepted key strokes.
        """

        if self.logging:
            QE_LOGGER.info(f"set_typing_delay(self, number={number})")

        config.typingDelay = number


    def set_debug_to_stdout_as(self, true_or_false=False):
        """
        Set debugging to stdout.

        @param true_or_false : bool
            Decision if debug to stdout or not.
        """

        if self.logging:
            QE_LOGGER.info(f"set_debug_to_stdout_as(self, true_or_false={true_or_false})")

        config.logDebugToStdOut = true_or_false


    def copy_data_folder(self):
        """
        Copy /data folder to /tmp/.
        """

        if self.logging:
            QE_LOGGER.info("copy_data_folder(self)")

        if os.path.isdir("data/"):
            for single_file in os.listdir("data/"):
                run(f"cp data/{single_file} /tmp/{single_file}")


    def detect_keyring(self):
        """
        Detect if keyring was setup. If not, setup the keyring with empty password.

        @param context : <context_object>
            Passed object.
        """

        if self.logging:
            QE_LOGGER.info("detect_keyring(self)")

        if not self.set_keyring:
            return

        current_user = os.path.expanduser("~")
        is_keyring_set = os.path.isfile("/tmp/keyring_set")
        is_keyring_in_place = os.path.isfile(f"{current_user}/.local/share/keyrings/default")

        if not is_keyring_set or not is_keyring_in_place:
            run(f"sudo rm -rf {current_user}/.local/share/keyrings/*")

            create_keyring_process = Popen("qecore_start_keyring", shell=True)
            self.keyring_process_pid = create_keyring_process.pid

            self.shell.child("Continue").click()
            self.shell.child("Continue").click()

            run(f"touch /tmp/keyring_set")

        run("killall qecore_start_keying")


    def capture_image(self):
        """
        Capture screenshot after failed step.
        """

        if self.logging:
            QE_LOGGER.info("capture_image(self)")

        if not self.production:
            return

        self.screenshot_run_result = run("gnome-screenshot -f /tmp/screenshot.png", verbose=True)


    def attach_screenshot_to_report(self, context):
        """
        Attach screenshot to report upon failed test.

        @param context : <context_object>
            Passed object.
        """

        if self.logging:
            QE_LOGGER.info("attach_screenshot_to_report(self, context)")

        if not self.production:
            return

        if self.screenshot_run_result[1] != 0:
            context.embed(mime_type="text/plain",
                          data=f"Screenshot capture failed: \n{self.screenshot_run_result}\n",
                          caption="Screenshot")
        else:
            data_base64 = base64.b64encode(open("/tmp/screenshot.png", "rb").read())
            data_encoded = data_base64.decode("utf-8").replace("\n", "")
            context.embed(mime_type="image/png",
                          data=data_encoded,
                          caption="Screenshot")


    def attach_image_to_report(self, context, image=None, caption="DefaultCaption"):
        """
        Attach image to report upon user request.

        @param context : <context_object>
            Passed object.
        @param image : str
            Location of image/png file.
        @param caption : str
            Caption that is to be displayed in test report.
        """

        if self.logging:
            QE_LOGGER.info("attach_image_to_report(self, context)")

        if not self.production:
            return

        if os.path.isfile(image):
            data_base64 = base64.b64encode(open(image, "rb").read())
            data_encoded = data_base64.decode("utf-8").replace("\n", "")
            context.embed(mime_type="image/png",
                          data=data_encoded,
                          caption=caption)


    def attach_video_to_report(self, context):
        """
        Attach video to report upon failed test.

        @param context : <context_object>
            Passed object.
        """

        if self.logging:
            QE_LOGGER.info("attach_video_to_report(self, context)")

        if not (self.production and self.record_video):
            return

        absolute_path_to_video = os.path.expanduser("~/Videos")
        screencast_list = [f"{absolute_path_to_video}/{file_name}" for file_name in \
            os.listdir(absolute_path_to_video) if "Screencast" in file_name]

        video_name = f"{self.component}_{self.current_scenario}"
        absolute_path_to_new_video = f"{absolute_path_to_video}/{video_name}.webm"

        if screencast_list == []:
            context.embed(mime_type="text/plain",
                          data="No video file found.",
                          caption="Video")
        else:
            if self.wait_for_stable_video:
                self.wait_for_video_encoding(screencast_list[0])

            data_base64 = base64.b64encode(open(screencast_list[0], "rb").read())
            data_encoded = data_base64.decode("utf-8").replace("\n", "")
            context.embed(mime_type="video/webm",
                          data=data_encoded,
                          caption="Video")
            run(f"mv {screencast_list[0]} {absolute_path_to_new_video}")
            run(f"sudo rm -rf {absolute_path_to_video}/Screencast*")


    def attach_journal_to_report(self, context):
        """
        Attach journal to report upon failed test.

        @param context : <context_object>
            Passed object.
        """

        if self.logging:
            QE_LOGGER.info("attach_journal_to_report(self, context)")

        if not self.production:
            return

        journal_run = run(" ".join((
            f"sudo journalctl",
            f"--all",
            f"--output=short-precise",
            f"--since='{self.logging_start}'"
            f"> /tmp/journalctl_short.log"
        )), verbose=True)

        if journal_run[1] != 0:
            context.embed(mime_type="text/plain",
                          data=f"Creation of journalctl file failed: \n{journal_run}\n",
                          caption="journalctl")
        else:
            context.embed(mime_type="text/plain",
                          data=open("/tmp/journalctl_short.log", "r").read(),
                          caption="journalctl")
            
        run("rm /tmp/journalctl_short.log")


    def attach_abrt_link_to_report(self, context):
        """
        Attach abrt link to report upon detected abrt FAF report.

        @param context : <context_object>
            Passed object.
        """

        if self.logging:
            QE_LOGGER.info("attach_abrt_link_to_report(self, context)")

        if not self.production:
            return

        faf_reports = set()
        abrt_directories = run("sudo ls /var/spool/abrt/ | grep ccpp-").strip("\n").split("\n")

        for abrt_directory in abrt_directories:
            try:
                reason_file = f"/var/spool/abrt/{abrt_directory}/reason"
                reported_to_file = f"/var/spool/abrt/{abrt_directory}/reported_to"

                abrt_faf_reason_run = run(f"sudo cat '{reason_file}'", verbose=True)
                abrt_faf_hyperlink_run = run(f"sudo cat '{reported_to_file}'", verbose=True)

                if abrt_faf_reason_run[1] == 0 and abrt_faf_hyperlink_run[1] == 0:
                    abrt_faf_reason = abrt_faf_reason_run[0].strip("\n")
                    abrt_faf_hyperlink = abrt_faf_hyperlink_run[0].split("ABRT Server: URL=")[-1].split("\n")[0]

                    faf_reports.add((abrt_faf_hyperlink, f"Reason: {abrt_faf_reason}"))

            except Exception as error:
                print(f"Exception caught: {error}")
                continue

        if faf_reports:
            context.embed("link", faf_reports, caption="FAF reports")


    def wait_for_video_encoding(self, file_name):
        """
        Wait until the video is fully encoded.
        This is verified by videos changing size.
        Once the file is encoded the size will not change anymore.

        @param file_name : str
            Video location for size verification.
        """

        if self.logging:
            QE_LOGGER.info(f"wait_for_video_encoding(self, file_name={file_name})")

        current_size = 0
        current_stability = 0

        try:
            with Timeout(15):
                while current_stability < 50:
                    new_size = os.path.getsize(file_name)
                    if current_size == new_size:
                        current_stability += 1
                    else:
                        current_stability = 0
                    current_size = new_size
                    sleep(0.1)

        except QETimeoutError:
            QE_LOGGER.info("Timeout reached while waiting for video encoding. Resuming embedding.")


    def set_background(self, color=None, background_image=None):
        """
        Change background to single color. Currently support for white/black/image.

        @param color : str
            String black/white to set as background color.
        @param background_image : str
            Image location to be set as background.
        """

        if self.logging:
            QE_LOGGER.info("".join((
                f"set_background(self, color={color}, ",
                f"background_image={background_image})."
            )))

        if background_image:
            run(f"gsettings set org.gnome.desktop.background picture-uri file://{background_image}")
        elif color == "white":
            run("gsettings set org.gnome.desktop.background picture-uri file://")
            run("gsettings set org.gnome.desktop.background primary-color \"#FFFFFF\"")
            run("gsettings set org.gnome.desktop.background secondary-color \"#FFFFFF\"")
            run("gsettings set org.gnome.desktop.background color-shading-type \"solid\"")
        elif color == "black":
            run("gsettings set org.gnome.desktop.background picture-uri file://")
            run("gsettings set org.gnome.desktop.background primary-color \"#000000\"")
            run("gsettings set org.gnome.desktop.background secondary-color \"#000000\"")
            run("gsettings set org.gnome.desktop.background color-shading-type \"solid\"")
        else:
            QE_LOGGER.info(" ".join((
                f"Color '{color}' is not defined.",
                f"You can define one yourself and submit merge request or"
                f"email modehnal@redhat.com for support."
            )))


    def overview_action(self, action="hide"):
        """
        Hide or show application overview.

        @param action : str
            Hide or show application overview.
        """

        if self.logging:
            QE_LOGGER.info(f"overview_action(self, action={action})")

        if action == "hide":
            run(" ".join((
                "dbus-send",
                "--session",
                "--type=method_call",
                "--dest=org.gnome.Shell",
                "/org/gnome/Shell",
                "org.gnome.Shell.Eval",
                "string:'Main.overview.hide();'"
            )))
        elif action == "show":
            run(" ".join((
                "dbus-send",
                "--session",
                "--type=method_call",
                "--dest=org.gnome.Shell",
                "/org/gnome/Shell",
                "org.gnome.Shell.FocusSearch"
            )))
        else:
            assert False, "Unknown option"



#!/usr/bin/env python3
from time import sleep
import shlex
import configparser
from subprocess import Popen, PIPE
from dogtail.rawinput import typeText, pressKey, keyCombo
from dogtail.tree import root
from qecore import run, QE_LOGGER
from qecore.time_limit import Timeout


class Application:
    def __init__(self, component, a11yAppName=None, desktopFileExists=True,
                 desktopFileName=None, appProcessName=None, shell=None,
                 session_type=None, session_desktop=None, logging=False):
        """
        Initiate Application instance.

        @param component : str
            Application name.
        @param a11yAppName : str
            Name of application as it appears in accessibility tree.
        @param desktopFileExists : bool
            States that desktop file of given application exists or not.
        @param desktopFileName : str
            Name of desktop file if the name is different from application name.
        @param appProcessName : str
            Name process of application if the name of is different from application name.
        @param logging : bool
            Turn on logging of this submodule. Passed from sandbox.
        """

        self.logging = logging

        if self.logging:
            QE_LOGGER.info(" ".join((
                f"__init__(self, component={component}, a11yAppName={a11yAppName}",
                f"desktopFileExists={desktopFileExists}, desktopFileName={desktopFileName}",
                f"appProcessName={appProcessName}, shell='expecting_passed_shell'",
                f"session_type={session_type}, session_desktop={session_desktop}",
                f"logging={str(logging)}"
            )))

        self.shell = shell
        self.session_type = session_type
        self.session_desktop = session_desktop
        self.pid = None
        self.instance = None
        self.component = component
        self.a11y_app_name = a11yAppName if a11yAppName else component
        self.desktop_file_exists = desktopFileExists
        self.exit_shortcut = "<Control_L><Q>"
        self.kill = True
        self.kill_command = None
        self.desktop_file_name = desktopFileName if desktopFileName else ""
        self.app_process_name = appProcessName if appProcessName else component
        self.get_desktop_file_data()


    def get_desktop_file_data(self):
        """
        Parse desktop file.
        """

        if self.logging:
            QE_LOGGER.info(f"{self.component} get_desktop_file_data(self)")

        if self.desktop_file_exists: # zenity/gnome-shell do not have desktop file
            desktop_run = run(" ".join((
                f"rpm -qlf $(which {self.component}) |",
                f"grep /usr/share/applications/ |",
                f"grep .desktop |",
                f"grep '{self.desktop_file_name}'"
            )), verbose=True)


            if desktop_run[1] != 0:
                raise Exception(f"Desktop file of application '{self.component}' was not found.")

            desktop_file = desktop_run[0].strip("\n")
            desktop_file_config = configparser.RawConfigParser()
            desktop_file_config.read(desktop_file)

            self.name = desktop_file_config.get("Desktop Entry", "name")
            self.exec = desktop_file_config.get("Desktop Entry", "exec").split(" ")[0]


    def start_via_command(self, command=None, in_session=False, kill=True):
        """
        Start application via command.

        @param command : str
            Complete command that is to be used to start application.
        @param in_session : bool
            Start application via command in session. TODO.
        @param kill : bool
            If application is running while command is run. Application will be killed.
        """

        if self.logging:
            QE_LOGGER.info("".join((
                f"{self.component} ",
                f"start_via_command(self, command={command}, in_session={in_session}, kill={kill})"
            )))

        if self.is_running() and kill:
            self.kill_application()

        if in_session:
            pressKey("Esc")
            keyCombo("<Alt><F2>")
            sleep(0.5)
            keyCombo("<Alt><F2>")
            sleep(0.5)
            enter_a_command = self.shell.findChild(lambda x: "activate" in x.actions and x.showing)
            enter_a_command.text = command if command else self.exec
            sleep(0.5)
            pressKey("Enter")
        else:
            self.process = Popen(shlex.split(command if command else self.exec),\
                 stdout=PIPE, stderr=PIPE)

        self.wait_before_app_starts(30)
        self.instance = self.get_root()


    def start_via_menu(self, kill=True):
        """
        Start application via command.

        @param kill : bool
            If application is running while start_via_menu is called. Application will be killed.
        """

        if self.logging:
            QE_LOGGER.info(f"{self.component} start_via_menu(self, kill={kill})")

        gnome_version_present = run("rpm -q gnome-shell").strip("\n")
        if ("classic" in self.session_desktop) and ("3.32" in gnome_version_present):
            assert False, f"Application cannot be started via menu in {gnome_version_present}."

        if not self.desktop_file_exists:
            raise Exception("".join((
                f"Testing target '{self.a11y_app_name}' does not have desktop file. ",
                f"Indication of user failure!"
            )))

        if self.is_running() and kill:
            self.kill_application()

        run(" ".join((
            "dbus-send",
            "--session",
            "--type=method_call",
            "--dest=org.gnome.Shell",
            "/org/gnome/Shell",
            "org.gnome.Shell.FocusSearch"
        )))

        sleep(1)
        typeText(self.name)
        pressKey("Enter")

        self.wait_before_app_starts(30)
        self.instance = self.get_root()


    def close_via_shortcut(self):
        """
        Start application via shortcut.
        """

        if self.logging:
            QE_LOGGER.info(f"{self.component} close_via_shortcut(self)")

        if not self.is_running():
            raise Exception("".join((
                f"Application '{self.a11y_app_name}' is no longer running. ",
                f"Indication of test failure!"
            )))

        keyCombo(self.exit_shortcut)

        self.wait_before_app_closes(30)
        self.instance = None


    def already_running(self):
        """
        If application is started by other means other than this submodule,
        this will initiate data that are necessary.
        """

        if self.logging:
            QE_LOGGER.info(f"{self.component} already_running(self)")

        self.wait_before_app_starts(15)
        self.instance = self.get_root()


    def get_root(self):
        """
        Get accessibility root of appllication.

        @return : <accessibility_object>
            Return root object of application.
        """

        if self.logging:
            QE_LOGGER.info(f"{self.component} get_root(self) search by '{self.a11y_app_name}'")

        return root.application(self.a11y_app_name)


    def is_running(self):
        """
        Get accessibility root of appllication.

        @return : bool
            Return state of application. Running or not.
        """

        if self.logging:
            QE_LOGGER.info(f"{self.component} is_running(self)")
        try:
            for _ in range(3):
                is_running = self.a11y_app_name in [x.name for x in root.applications()]
                if is_running:
                    break
            return is_running
        except Exception:
            return False


    def get_pid_list(self):
        """
        Get list of processes of running application.

        @return : str
            Return all running processes of application, seperated by new line character,
            not converting to list. Return nothing if application is not running.
        """

        if self.logging:
            QE_LOGGER.info(f"{self.component} get_pid_list(self)")

        if self.is_running():
            return run(f"pgrep -fla {self.app_process_name}").strip("\n")

        return None


    def get_all_kill_candidates(self):
        """
        Take result of get_pid_list and return only processes of application.
        If kill candidate happens to have ['runtest', 'behave', 'dogtail', '/usr/bin/gnome-shell']
        in its process name. Process will not be killed.
        This prevents testname colliding with found process so we will not kill the test itself.

        @return : list
            Return all processed id of applications.
            Return empty list if no satisfactory process was found.
        """

        if self.logging:
            QE_LOGGER.info(f"{self.component} get_all_kill_candidates(self)")

        application_pid_string = self.get_pid_list()
        if application_pid_string:
            application_pid_list = application_pid_string.split("\n")
        else:
            return []

        final_pid_list = []
        for process in application_pid_list:
            if not (("runtest" in process) or \
                    ("behave" in process) or \
                    ("dogtail" in process) or \
                    (process == "/usr/bin/gnome-shell")):
                extracted_pid = process.split(" ", 1)[0]
                try:
                    final_pid_list.append(int(extracted_pid))
                except ValueError:
                    pass # skip non-digits
        return final_pid_list


    def kill_application(self):
        """
        Kill application.
        """

        if self.logging:
            QE_LOGGER.info(f"{self.component} kill(self)")

        if self.is_running() and self.kill:
            if not self.kill_command:
                for pid in self.get_all_kill_candidates():
                    run(f"sudo kill -9 {pid} > /dev/null")
            else:
                run(self.kill_command)


    # Following two could be merged, I could not think of a nice way of doing it though.
    def wait_before_app_starts(self, time_limit):
        """
        Wait before application starts.

        @param time_limit : int
            Number which signifies time before the run is stopped.
        """

        if self.logging:
            QE_LOGGER.info("".join((
                f"{self.component} ",
                f"wait_before_app_starts(self, time_limit={time_limit})"
            )))

        with Timeout(time_limit):
            try:
                while not self.is_running():
                    sleep(0.1)
            except Exception:
                assert False, "".join((
                    f"Application '{self.a11y_app_name}' is not running. ",
                    f"Indication of test failure!"
                ))

    def wait_before_app_closes(self, time_limit):
        """
        Wait before application stops.

        @param time_limit : int
            Number which signifies time before the run is stopped.
        """

        if self.logging:
            QE_LOGGER.info("".join((
                f"{self.component} ",
                f"wait_before_app_closes(self, time_limit={time_limit}"
            )))

        with Timeout(time_limit):
            try:
                while self.is_running():
                    sleep(0.1)
            except Exception:
                assert False, "".join((
                    f"Application '{self.a11y_app_name}' is running. ",
                    f"Indication of test failure!"
                ))



#!/usr/bin/env python3
import sys
from os import path
from time import sleep
from dogtail.rawinput import click
from behave import step
from qecore import run

try:
    import cv2
except ModuleNotFoundError:
    print("You need to install an 'opencv-python' via pip or 'python3-opencv' via yum/dnf.")
    sys.exit(1)

HELP_MESSAGE = """
You are encouraged to build your own step function according to your needs.
Two steps that you see bellow are:
    * General step that just compares and asserts the result.
    * General step that just compares and clicks on the found result.

What is needed for image match:
    * You need to capture an image in which we look for the element you want to find.
        * Provided by capture_image method in Matcher class.
        * This option is True by default.
        * If you have your own, set capture=False and provide self.screen_path in the Matcher class.

    * You need to match the two images, you are looking for a 'needle'.
      So you provide it in function or in step call (.feature file).
        * Provided by match which will return True or False. Lets user react on False return value.
        * Provided by assert_match which will assert the result and terminate the test on False.

    * (Optional) You can draw the result for attachment or
            your own confirmation that matching works.
        * Provided by draw method on Matcher instance to get an image with highlighted needle.
          Highlight is a red rectangle exactly in a place of a match, surrounding provided needle.

    * (Optional) You can click on your found result.
        * Provided by click method in Matcher instance.
        * Requirements are of course success of a match/assert_match.

    * (Optional) You can embed result to test report.
        * For this option the method draw is required.
        * Use method provided in TestSandbox class:
            attach_image_to_report(context, image=image_location, caption="DefaultCaption")
        * Or embed it on your own:
            context.embed("image/png", open(image_location, "rb").read(), caption="DefaultCaption")
        * Remember that result is saved in Matcher instance as
            self.diff_path which equals "/tmp/diff.png"
"""

@step('Image "{needle}" is shown on screen')
@step('Image "{needle}" is shown on screen with threshold "{threshold:d}"')
def image_match(context, needle, threshold=0.8):
    image_match_instance = Matcher()
    image_match_instance.assert_match(needle, threshold)
    image_match_instance.draw()


@step('Locate and click "{needle}"')
def locate_and_click(context, needle):
    image_match_instance = Matcher()
    image_match_instance.assert_match(needle)
    image_match_instance.click()


class Matcher:
    def __init__(self):
        self.screen_path = "/tmp/pic.png"
        self.diff_path = "/tmp/diff.png"
        self.capture_image_cmd = f"gnome-screenshot -f {self.screen_path}"
        self.needle_width = 0
        self.needle_height = 0
        self.matched_value = 0.0
        self.matched_loc = (0, 0)

        self.ori_img = None
        self.ori_img_gray = None
        self.needle = None
        self.needle_size = None

    def capture_image(self):
        run(self.capture_image_cmd)


    def assert_match(self, needle, threshold=0.8, capture=True):
        assert self.match(needle, threshold, capture), \
            f"Image match value: {self.matched_value}"


    def match(self, needle, threshold=0.8, capture=True):
        if capture:
            self.capture_image()

        self.ori_img = cv2.imread(self.screen_path)
        self.ori_img_gray = cv2.cvtColor(self.ori_img, cv2.COLOR_BGR2GRAY)
        self.needle = cv2.imread(path.abspath(needle), 0)
        self.needle_width, self.needle_height = self.needle.shape[::-1]

        match = cv2.matchTemplate(self.ori_img_gray, self.needle, cv2.TM_CCOEFF_NORMED)
        _, self.matched_value, _, self.matched_loc = cv2.minMaxLoc(match)

        return self.matched_value > threshold


    def draw(self):
        self.needle_size = (self.matched_loc[0] + self.needle_width, \
                            self.matched_loc[1] + self.needle_height)
        cv2.rectangle(self.ori_img, self.matched_loc, self.needle_size, (0, 0, 255), 2)
        cv2.imwrite(self.diff_path, self.ori_img)


    def click(self):
        match_center_x = self.matched_loc[0] + int(self.needle_width / 2)
        match_center_y = self.matched_loc[1] + int(self.needle_height / 2)
        click(match_center_x, match_center_y)
        sleep(1)



#!/usr/bin/env python3
import os
import shlex
import types
import logging
from time import sleep, time
from subprocess import CalledProcessError, check_output, STDOUT
from functools import partial
from dogtail.rawinput import holdKey, releaseKey


__author__ = """Michal Odehnal <modehnal@redhat.com>"""
__version__ = "1.0.0"
__copyright__ = "Copyright © 2018-2020 Red Hat, Inc."
__license__ = "GPL"
__all__ = ("application", "common_steps", "get_node", "image_matching", "sandbox", "time_limit")


QE_LOGGER = logging.getLogger("DESKTOPQE")
FORMAT = "[%(filename)s:%(lineno)3s] %(message)s"
logging.basicConfig(format=FORMAT)
QE_LOGGER.setLevel(logging.INFO)

QE_DEVELOPMENT = not os.path.isdir("/mnt/tests/")


def get_application(context, application):
    """
    Get <accessibility_object> of application root, based upon given name.

    @param context : <context_object>
        Context object that is passed from common steps.
    @param application : str
        String of application identification: name.
    @return : <accessibility_object>
        Return root object of application.
    """

    app_class_to_return = None
    try:
        app_class_to_return = getattr(context, application)
    except AttributeError:
        for app in context.sandbox.applications:
            if app.component == application:
                app_class_to_return = app
                break
    except TypeError:
        app_class_to_return = context.sandbox.default_application
        assert context.sandbox.default_application is not None, \
            "Default application was not found. Check your environment file!"

    assert app_class_to_return is not None, \
        "Application was not found. Check your environment or feature file!"
    assert not isinstance(app_class_to_return, str), \
        "Application class was not found. Usually indication of not installed application."

    return app_class_to_return


def run(command, verbose=False):
    """
    Execute a command and get output.

    @param command : str
        Command to be executed.
    @param verbose : bool
        Boolean value for verbose option.
    @return : str
        Return string value of command output.
    @return : list
        Return list value of command upon verbose option \
        set in format (output, return code, exception).
    """

    try:
        output = check_output(command, shell=True, stderr=STDOUT, encoding="utf-8")
        return output if not verbose else (output, 0, None)
    except CalledProcessError as error:
        return error.output if not verbose else (error.output, error.returncode, error)

#behave-common-steps leftover
def wait_until(tested, element=None, timeout=30, period=0.25, params_in_list=False):
    """
    This function keeps running lambda with specified params until the
    result is True or timeout is reached. Instead of lambda, Boolean variable
    can be used instead.
    Sample usages:
     * wait_until(lambda x: x.name != 'Loading...', context.app.instance)
       Pause until window title is not 'Loading...'.
       Return False if window title is still 'Loading...'
       Throw an exception if window doesn't exist after default timeout

     * wait_until(lambda element, expected: x.text == expected,
           (element, 'Expected text'), params_in_list=True)
       Wait until element text becomes the expected (passed to the lambda)

     * wait_until(dialog.dead)
       Wait until the dialog is dead

    """

    if isinstance(tested, bool):
        curried_func = lambda: tested
    # or if callable(tested) and element is a list or a tuple
    elif isinstance(tested, types.FunctionType) and \
            isinstance(element, (tuple, list)) and params_in_list:
        curried_func = partial(tested, *element)
    # or if callable(tested) and element is not None?
    elif isinstance(tested, types.FunctionType) and element is not None:
        curried_func = partial(tested, element)
    else:
        curried_func = tested

    exception_thrown = None
    mustend = int(time()) + timeout
    while int(time()) < mustend:
        try:
            if curried_func():
                return True
        except Exception as error:  # pylint: disable=broad-except
            # If lambda has thrown the exception we'll re-raise it later
            # and forget about if lambda passes
            exception_thrown = error
        sleep(period)
    if exception_thrown is not None:
        raise exception_thrown  # pylint: disable=raising-bad-type
    else:
        return False


KEY_VALUE = {
    "Alt" : 64, "Alt_L" : 64, "Alt_R" : 108,
    "Shift" : 50, "Shift_L" : 50, "Shift_R" : 62,
    "Ctrl" : 37, "Tab" : 23, "Super" : 133,
}

class HoldKey:
    def __init__(self, key_name):
        self.key_name = key_name
        holdKey(self.key_name)
    def __enter__(self):
        return self
    def __exit__(self, my_type, value, traceback):
        releaseKey(self.key_name)


class Tuple(tuple):
    def __add__(self, other):
        return Tuple(x + y for x, y in zip(self, other))
    def __rmul__(self, other):
        return Tuple(other * x for x in self)
    def __eq__(self, other):
        return (x == y for x, y in zip(self, other))
    def __lt__(self, other):
        return self[0] < other[0] or self[1] < other[1]
    def __gt__(self, other):
        return self[0] > other[0] or self[1] > other[1]


def validate_command(command):
    # Lets take care of any scripts user would like to try. Test
    parsed_command = shlex.split(command)
    valid_command = ""
    for command_part in parsed_command:
        for character in command_part:
            valid_command += f"\\{character}" if not character.isalpha() else character
        valid_command += " "
    return valid_command


def verify_path(template_path):
    try:
        assert os.path.isfile(template_path)
    except Exception as error:
        assert False, f"Desired file was not found: {error}"
    return template_path


SPACER = ' '
def plain_dump(node):
    def crawl(node, depth):
        dump(node, depth)
        for child in node.children:
            crawl(child, depth + 1)

    def dump_std_out(item, depth):
        # str wont possibly work in p3
        print(SPACER * depth + str(item) + \
            f"     [p:{item.position}, s:{item.size}, vis:{item.visible}, show:{item.showing}]")

    dump = dump_std_out
    crawl(node, 0)



#!/usr/bin/env python3
import traceback
import sys
from dogtail.tree import root
from qecore import QE_DEVELOPMENT

if QE_DEVELOPMENT:
    try:
        from termcolor import colored
    except ModuleNotFoundError:
        QE_DEVELOPMENT = False


class GetNode:
    def __init__(self, context, name, role_name, description,
                 m_btn, attr, a11y_root_name, retry, expect_positive):
        """
        Initiate GetNode instance.

        @param context : <context_object>
            Context object that is passed from common steps.
        @param name : str
            Node identification: name.
        @param role_name : str
            Node identification: roleName.
        @param description : str
            Node identification: description.
        @param m_btn : str
            Mouse click after node identification.
            Accepted values are "Left" and "Right".
        @param attr : str
            Node identification: attribute.
            The most used options are: ["showing", "visible", "checked", "focused", "sensitive"]
        @param intance_root_name : str
            Application name.
            Application name to be found in context.<app_name> or in accessibility tree.
            If search of accessibility tree fails the context \
                object will be examined for Application instance.
        @param retry : bool
            Option to give search function to look again a few times if search fails.
            Used for slower applications. User might want to \
                click right away but window can have a few secods delay.
        @param expect_positive : bool
            Option to pass the common step call if the node is not found.
            Some steps might want the node not to be found.
        """
        a11y_roots = {x.name: x for x in root.applications()}

        if a11y_root_name in a11y_roots:
            self.root = a11y_roots[a11y_root_name]
        elif a11y_root_name is not None and hasattr(context, a11y_root_name):
            self.root = getattr(context, a11y_root_name).instance
        else:
            self.root = None

        if self.root is None:
            try:
                self.root = context.sandbox.default_application.instance
            except AttributeError:
                assert False, "".join((
                    "You need to define a default application ",
                    "if you are using steps without root."
                ))

        if m_btn == "Left":
            self.m_btn = 1
        elif m_btn == "Middle":
            self.m_btn = 2
        elif m_btn == "Right":
            self.m_btn = 3
        elif m_btn is None:
            self.m_btn = None
        else:
            self.m_btn = 0

        assert self.m_btn != 0, "Unknown mouse button! Check your feature file!"

        self.name = ("".join(name) if name != "Empty" else "") \
            if name != "None" else None
        self.role_name = ("".join(role_name) if role_name != "Empty" else "") \
            if role_name != "None" else None
        self.description = None if description is None else "".join(description) \
            if description != "Empty" else ""

        self.attr = attr if attr in ["showing", "visible", "checked", "focused", "sensitive"] \
            else None if attr is None else False
        assert self.attr is not False, "Unknown attribute. Check your feature file!"

        self.retry = retry if isinstance(retry, bool) else None
        assert self.retry is not None, "Unknown retry state. Check your feature file!"

        self.expect_positive = expect_positive if isinstance(expect_positive, bool) else None
        assert self.expect_positive is not None, "".join((
            f"Unknown expect_positive state: {self.expect_positive}. ",
            f"Check your feature file!"
        ))


    def __enter__(self):
        try:
            found_node = self.root.findChild(lambda x: \
                ((not self.name is not None) or self.name in repr(x.name)) and \
                ((not self.role_name is not None) or self.role_name == x.roleName) and \
                ((not self.description is not None) or self.description in x.description) and \
                ((not self.attr is not None) or getattr(x, self.attr)) and \
                x.position[0] >= 0 and \
                x.size[0] > 0, \
                retry=self.retry)
        except Exception as error:
            if self.expect_positive:
                assert False, get_error_message(self, error)
            else:
                found_node = None
        return (self, found_node)


    def __exit__(self, exc_type, exc_value, trcb):
        if exc_type is not None:
            traceback.print_exc(file=sys.stdout)
            return False
        return True


def get_center(node):
    return (node.position[0] + node.size[0]/2, node.position[1] + node.size[1]/2)


def get_formated_duplicates(list_size, list_of_duplicates, attr):
    return "".join(sorted(set(["\t{0}: '{1}' {2}: '{3}' {4}: '{5}' {6}: '{7}' {8}: '{9}'\n".format(
        colored("name", "yellow") if QE_DEVELOPMENT else "name", repr(node.name),
        colored("roleName", "yellow") if QE_DEVELOPMENT else "roleName", node.roleName,
        colored("position", "yellow") if QE_DEVELOPMENT else "position", node.position,
        colored("size", "yellow") if QE_DEVELOPMENT else "size", node.size,
        colored(f"{attr}", "yellow") if QE_DEVELOPMENT else f"{attr}" \
            if attr else "attribute", getattr(node, attr) if attr else "None") \
    for node in list_of_duplicates]), key=str.lower)) if list_size != 0 else "\tNone\n"


def get_formated_error_message(error,
                               node_name, same_name_items,
                               node_role_name, same_role_name_items):
    return "".join(["\n\n{0}: {1}\n\n{2}: {3}:\n{4}\n{5}: {6}:\n{7}\n".format(
        colored("Item was not found", "yellow", attrs=["bold"]) \
            if QE_DEVELOPMENT else "Item was not found", error,
        colored("Items with name", "yellow", attrs=["bold"]) \
            if QE_DEVELOPMENT else "Items with name", repr(node_name), same_name_items,
        colored("Items with roleName", "yellow", attrs=["bold"]) \
            if QE_DEVELOPMENT else "Items with roleName", node_role_name, same_role_name_items)])


def get_error_message(node, error):
    nodes_with_name = node.root.findChildren(lambda x: \
        node.name in x.name and (not (node.name != "") or x.name != ""))
    nodes_with_name_size = len(nodes_with_name)
    nodes_with_name_formatted = get_formated_duplicates(nodes_with_name_size,
                                                        nodes_with_name,
                                                        node.attr)

    nodes_with_role_name = node.root.findChildren(lambda x: x.roleName == node.role_name)
    nodes_with_role_name_size = len(nodes_with_role_name)
    nodes_with_role_name_formatted = get_formated_duplicates(nodes_with_role_name_size,
                                                             nodes_with_role_name,
                                                             node.attr)

    return get_formated_error_message(error,
                                      node.name,
                                      nodes_with_name_formatted,
                                      node.role_name,
                                      nodes_with_role_name_formatted)



#!/usr/bin/env python3
from time import sleep
from dogtail.rawinput import typeText, pressKey, keyCombo, absoluteMotion, click
from behave import step
from qecore import get_application
from qecore.get_node import GetNode, get_center


@step('{m_btn} click "{name}" "{role_name}"')
@step('{m_btn} click "{name}" "{role_name}" in "{root}"')
@step('{m_btn} click "{name}" "{role_name}" that is "{attr}"')
@step('{m_btn} click "{name}" "{role_name}" that is "{attr}" in "{root}"')
@step('{m_btn} click "{name}" "{role_name}" with description "{descr}"')
@step('{m_btn} click "{name}" "{role_name}" with description "{descr}" in "{root}"')
@step('{m_btn} click "{name}" "{role_name}" with description "{descr}" that is "{attr}"')
@step('{m_btn} click "{name}" "{role_name}" with description "{descr}" that is "{attr}" in "{root}"')
def mouse_click(context, name=None, role_name=None, descr=None, m_btn=None, attr=None, root=None, retry=True, expect_positive=True):
    with GetNode(context, name, role_name, descr, m_btn, attr, root, retry, expect_positive) as (data, node):
        node.click(data.m_btn)


@step('Mouse over "{name}" "{role_name}"')
@step('Mouse over "{name}" "{role_name}" in "{root}"')
@step('Mouse over "{name}" "{role_name}" that is "{attr}"')
@step('Mouse over "{name}" "{role_name}" that is "{attr}" in "{root}"')
@step('Mouse over "{name}" "{role_name}" with description "{descr}"')
@step('Mouse over "{name}" "{role_name}" with description "{descr}" in "{root}"')
@step('Mouse over "{name}" "{role_name}" with description "{descr}" that is "{attr}"')
@step('Mouse over "{name}" "{role_name}" with description "{descr}" that is "{attr}" in "{root}"')
def mouse_over(context, name=None, role_name=None, descr=None, m_btn=None, attr=None, root=None, retry=True, expect_positive=True):
    with GetNode(context, name, role_name, descr, m_btn, attr, root, retry, expect_positive) as (_, node):
        absoluteMotion(*(get_center(node)))


@step('Make an action "{action}" for "{name}" "{role_name}"')
@step('Make an action "{action}" for "{name}" "{role_name}" in "{root}"')
@step('Make an action "{action}" for "{name}" "{role_name}" that is "{attr}"')
@step('Make an action "{action}" for "{name}" "{role_name}" that is "{attr}" in "{root}"')
@step('Make an action "{action}" for "{name}" "{role_name}" with description "{descr}"')
@step('Make an action "{action}" for "{name}" "{role_name}" with description "{descr}" in "{root}"')
@step('Make an action "{action}" for "{name}" "{role_name}" with description "{descr}" that is "{attr}"')
@step('Make an action "{action}" for "{name}" "{role_name}" with description "{descr}" that is "{attr}" in "{root}"')
def make_action(context, action=None, name=None, role_name=None, descr=None, m_btn=None, attr=None, root=None, retry=True, expect_positive=True):
    with GetNode(context, name, role_name, descr, m_btn, attr, root, retry, expect_positive) as (_, node):
        node.doActionNamed(action)


@step('Item "{name}" "{role_name}" found')
@step('Item "{name}" "{role_name}" found in "{root}"')
@step('Item "{name}" "{role_name}" is "{attr}"')
@step('Item "{name}" "{role_name}" is "{attr}" in "{root}"')
@step('Item "{name}" "{role_name}" with description "{descr}" is "{attr}"')
@step('Item "{name}" "{role_name}" with description "{descr}" is "{attr}" in "{root}"')
def node_attribute(context, name=None, role_name=None, descr=None, m_btn=None, attr=None, root=None, retry=True, expect_positive=True):
    with GetNode(context, name, role_name, descr, m_btn, attr, root, retry, expect_positive) as (_, node):
        assert node is not None, "Node was not found, it should be!"


@step('Item "{name}" "{role_name}" was not found')
@step('Item "{name}" "{role_name}" was not found in "{root}"')
@step('Item "{name}" "{role_name}" is not "{attr}"')
@step('Item "{name}" "{role_name}" is not "{attr}" in "{root}"')
@step('Item "{name}" "{role_name}" with description "{descr}" is not "{attr}"')
@step('Item "{name}" "{role_name}" with description "{descr}" is not "{attr}" in "{root}"')
def node_not_attribute(context, name=None, role_name=None, descr=None, m_btn=None, attr=None, root=None, retry=True, expect_positive=False):
    with GetNode(context, name, role_name, descr, m_btn, attr, root, retry, expect_positive) as (_, node):
        assert node is None, "Node was found, it should not be!"


@step('Item "{name}" "{role_name}" has text "{text}"')
@step('Item "{name}" "{role_name}" has text "{text}" in "{root}"')
@step('Item "{name}" "{role_name}" with description "{descr}" has text "{text}"')
@step('Item "{name}" "{role_name}" with description "{descr}" has text "{text}" in "{root}"')
def node_with_text(context, name=None, role_name=None, descr=None, text=None, m_btn=None, attr=None, root=None, retry=True, expect_positive=True):
    with GetNode(context, name, role_name, descr, m_btn, attr, root, retry, expect_positive) as (_, node):
        assert text in node.text, "".join((
            f"Found node should have text: {text}\n",
            f"Instead the node has text: {node.text}"
        ))


@step('Item "{name}" "{role_name}" does not have text "{text}"')
@step('Item "{name}" "{role_name}" does not have text "{text}" in "{root}"')
@step('Item "{name}" "{role_name}" with description "{descr}" does not have text "{text}"')
@step('Item "{name}" "{role_name}" with description "{descr}" does not have text "{text}" in "{root}"')
def node_without_text(context, name=None, role_name=None, descr=None, text=None, m_btn=None, attr=None, root=None, retry=True, expect_positive=True):
    with GetNode(context, name, role_name, descr, m_btn, attr, root, retry, expect_positive) as (_, node):
        assert not text in node.text, "".join((
            f"Found node should have text: {text}\n",
            f"Node was found with text: {node.text}"
        ))


@step('Item "{name}" "{role_name}" does not have description "{descr}"')
@step('Item "{name}" "{role_name}" does not have description "{descr}" in "{root}"')
@step('Item "{name}" "{role_name}" does not have description "{descr}" that is "{attr}"')
@step('Item "{name}" "{role_name}" does not have description "{descr}" that is "{attr}" in "{root}"')
def node_without_description(context, name=None, role_name=None, descr=None, m_btn=None, attr=None, root=None, retry=True, expect_positive=True):
    with GetNode(context, name, role_name, None, m_btn, attr, root, retry, expect_positive) as (_, node):
        assert not descr in node.description, "".join((
            f"Found node should not have description: {descr}\n",
            f"Instead the node has description: {node.description}"
        ))


@step('Wait until "{name}" "{role_name}" is "{attr}"')
@step('Wait until "{name}" "{role_name}" is "{attr}" in "{root}"')
@step('Wait until "{name}" "{role_name}" with description "{descr}" is "{attr}"')
@step('Wait until "{name}" "{role_name}" with description "{descr}" is "{attr}" in "{root}"')
def wait_until_attr(context, name=None, role_name=None, descr=None, m_btn=None, attr=None, root=None, retry=True, expect_positive=True):
    with GetNode(context, name, role_name, descr, m_btn, attr, root, retry, expect_positive) as (_, node):
        time_went_by = 0
        while not node.sensitive and time_went_by < 15:
            time_went_by += 0.5
            sleep(0.5)


#@step('Start another instance of "{application}" via "{start_via}"') # to be solved
#@step('Start another instance of "{application}" via command:"{command}"') # to be solved
@step('Start {application} via {start_via}') # behave-common-steps decorator
@step('Start "{application}" via command in {session}') # behave-common-steps decorator
@step('Start application "{application}" via "{start_via}"')
@step('Start application "{application}" via command "{command}"')
@step('Start application "{application}" via command in "{session}"')
@step('Start application "{application}" via command "{command}" in "{session}"')
def start_application(context, application=None, start_via="command", command=None, session=None):
    application = get_application(context, application)
    if start_via == "menu":
        try:
            application.start_via_menu()
        except Exception:
            application.start_via_menu()
    elif start_via == "command":
        try:
            application.start_via_command(command=command, in_session=session)
        except Exception:
            application.start_via_command(command=command, in_session=session)
    else:
        raise AssertionError("Only defined options are 'command' and 'menu'.")


@step('Close app via gnome panel') # behave-common-steps decorator
@step('Close application "{application}" via "{close_via}"')
def application_in_not_running(context, application=None, close_via="gnome panel"):
    application = get_application(context, application)
    if close_via == "gnome panel":
        context.execute_steps(f'* Left click "{application.name}" "menu" in "gnome-shell"')
        sleep(0.5)
        context.execute_steps('* Left click "Quit" "label" in "gnome-shell"')

    elif close_via == "application menu":
        application.instance.children[0][0].click(3)
        context.execute_steps(f'* Left click "Close" "label" in "gnome-shell"')

    elif close_via == "shortcut":
        application.close_via_shortcut()

    elif close_via == "kill command":
        application.kill_application()

    else:
        raise AssertionError("".join((
            "Only defined options are:\n",
            "'gnome panel', 'application menu', 'shortcut' and 'kill command'."
        )))


@step('{application} shouldn\'t be running anymore') # behave-common-steps decorator
@step('Application "{application}" is no longer running')
def application_is_not_running(context, application):
    application = get_application(context, application)
    if application.is_running():
        application.wait_before_app_closes(15)


@step('{application} should start') # behave-common-steps decorator
@step('Application "{application}" is running')
def application_is_running(context, application):
    application = get_application(context, application)
    application.already_running()
    if not application.is_running():
        application.wait_before_app_starts(15)


@step('Click "{target_name}" in GApplication menu') # behave-common-steps decorator
def click_gapp_menu(context, target_name):
    assert context.sandbox.default_application is not None, \
        "You need to define a default application if you are using steps without root."
    context.execute_steps(f'* Left click "{context.sandbox.default_application.name}" "menu" in "gnome-shell"')
    sleep(0.5)
    context.execute_steps(f'* Left click "{target_name}" "label" in "gnome-shell"')


@step('Type text: "{text}"')
def type_text(context, text):
    typeText(text)


@step('Press key: "{key_name}"')
def press_key(context, key_name):
    pressKey(key_name)


@step('Press "{combo_name}"') # behave-common-steps decorator
@step('Key combo: "{combo_name}"')
def key_combo(context, combo_name):
    keyCombo(combo_name)


@step('Wait {number} second before action')
@step('Wait {number} seconds before action')
def wait_up(context, number):
    sleep(int(number))


@step('Move mouse to: x: "{position_x}", y: "{position_y}"')
def absolutie_motion(context, position_x, position_y):
    absoluteMotion(int(position_x), int(position_y))


@step('{button} click on: x: "{position_x}", y: "{position_y}"')
def click_on_position(context, button, position_x, position_y):
    buttons = dict(Left=1, Middle=2, Right=3)
    click(int(position_x), int(position_y), buttons[button])



#!/usr/bin/env python3
import setuptools

def scripts():
    import os
    scripts_list = os.listdir(os.curdir + '/scripts/')
    result = []
    for file in scripts_list:
        result = result + ['scripts/' + file]
    return result


setuptools.setup(
    name="qecore",
    version="1.0.5",
    author="Michal Odehnal",
    author_email="modehnal@redhat.com",
    description="DesktopQE Tool for unified test execution",
    url="",
    packages=setuptools.find_packages(),
    scripts=scripts(),
    install_requires=["behave",
                      "behave-html-formatter",
                      "pydbus",
                      "termcolor"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)
